import logging
import socket
import re
import _thread
import time
import urllib3

from multiprocessing.pool import ThreadPool
from ruxit.api.base_plugin import RemoteBasePlugin

logger = logging.getLogger(__name__)

class CustomNetScalerPluginRemote(RemoteBasePlugin):
    # Used for threading
    start = 0
    
    # Used for limiting events
    MAX_EVENTS = 99
    events = 0

    # Used for not spamming events
    previousEvents = []
    raisedEvents = []
    
    # For adding endpoints
    listeningPorts = set()
    listeningIPs = set()
    
    # Used to always have debug logging enabled the first run
    firstRun = True
    
    # Used to check if a threads was successfully polled
    successfullyPolled = False
    
    # Used for metric dimension limitations
    MAX_SPLITS = 1000

    session = False
    
    api = None
    
    def terminate(self, **kwargs):
        if self.api != None:
            self.api.logout()

    def query(self, **kwargs):
        self.listeningPorts = set()
        self.listeningIPs = set()
        self.events = 0
        self.raisedEvents = []
        self.start = time.time()
        config = kwargs["config"]
        username = config["auth_user"]
        password = config["auth_password"]
        url = config["host_name"]
        
        #create verify request parameter
        verifyCertificate = config["certificate"]
        certificatePath = config["certificatePath"].strip()
        verify = verifyCertificate
        if verify and certificatePath != "":
            verify = certificatePath
            
        self.session = False
        if "session" in config:
            self.session = config["session"]
            
        debugLogging = config["debug"]
        group = config["group"]
        if debugLogging or self.firstRun:
            logger.setLevel(logging.DEBUG)
            self.firstRun = False
        else:
            logger.setLevel(logging.WARNING)
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            
        regexLBVirtualServers = []
        pollLBVirtualServers = True
        if "VirtualServers" in config and config["VirtualServers"] != "":
            if config["VirtualServers"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollLBVirtualServers = False
                logger.info("Not polling load balancing virtual servers")
            else:
                for lbVirtualServer in config["VirtualServers"].split(','):
                    regexLBVirtualServers.append(re.compile(lbVirtualServer + "$"))
                    logger.info("Adding lb virtual server filter: " + lbVirtualServer)
            
        regexCSVirtualServers = []
        pollCSVirtualServers = True
        if "csVirtualServers" in config and config["csVirtualServers"] != "":
            if config["csVirtualServers"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollCSVirtualServers = False
                logger.info("Not polling content switching virtual servers")
            else:
                for csVirtualServer in config["csVirtualServers"].split(','):
                    regexCSVirtualServers.append(re.compile(csVirtualServer + "$"))
                    logger.info("Adding content switching virtual server filter: " + csVirtualServer)
            
        regexVPNVirtualServers = []
        pollVPNVirtualServers = True
        if "vpnVirtualServers" in config and config["vpnVirtualServers"] != "":
            if config["vpnVirtualServers"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollVPNVirtualServers = False
                logger.info("Not polling VPN virtual servers")
            else:
                for vpnVirtualServer in config["vpnVirtualServers"].split(','):
                    regexVPNVirtualServers.append(re.compile(vpnVirtualServer + "$"))
                    logger.info("Adding vpn virtual server filter: " + vpnVirtualServer)
            
        regexServices = []
        pollServices = True
        if "services" in config and config["services"] != "":
            if config["services"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollServices = False
                logger.info("Not polling services")
            else:
                for service in config["services"].split(','):
                    regexServices.append(re.compile(service + "$"))
                    logger.info("Adding service filter: " + service)
            
        regexServiceGroups = []
        pollServiceGroups = True
        if "servicegroups" in config and config["servicegroups"] != "":
            if config["servicegroups"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollServiceGroups = False
                logger.info("Not polling service groups")
            else:
                for servicegroup in config["servicegroups"].split(','):
                    regexServiceGroups.append(re.compile(servicegroup + "$"))
                    logger.info("Adding service group filter: " + servicegroup)
            
        regexInterfaces = []
        pollInterfaces = True
        if "interfaces" in config and config["interfaces"] != "":
            if config["interfaces"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollInterfaces = False
                logger.info("Not polling interfaces")
            else:
                for interface in config["interfaces"].split(','):
                    regexInterfaces.append(re.compile(interface + "$"))
                    logger.info("Adding interface filter: " + service)
            
        hostName = url.replace("http://","").replace("https://","")
        hostName = hostName.split(':')[0].split('/')[0]
        g1 = self.topology_builder.create_group("NetScaler - " + group, "NetScaler - " + group)
        e1 = g1.create_element("NetScaler - " + group + " - " + hostName, "NetScaler - " + group + " - " + hostName)
        e1.report_property("Plugin version", kwargs["json_config"]["version"])
        
        # All the endpoints currently polled
        # printName = name used in events
        # url = url called to get the data
        # nameAttr = name of the endpoint used for splitting, leave out if there is no splitting
        # splitName = Splitting dimension
        # limitSplitting = only include splitting if it matches a regex in the passed array
        # metrics = The metrics fetched from the statistics of the endpoint
        endpoints = []
        system = {"printName": "System", "url": "/stat/system", "rootElement": "system", "metrics": []}
        endpoints.append(system)
        aaa = {"printName": "aaa", "url": "/stat/aaa", "rootElement": "aaa", "metrics": []}
        endpoints.append(aaa)
        compression = {"printName": "Compression", "url": "/stat/cmp", "rootElement": "cmp", "metrics": []}
        endpoints.append(compression)
        lbvirtualServer = {"printName": "Load balancing virtual server", "url": "/stat/lbvserver", "rootElement": "lbvserver", "splitName": "VirtualServer", "limitSplitting": regexLBVirtualServers, "nameAttr": "name", "metrics": []}
        if pollLBVirtualServers:
            endpoints.append(lbvirtualServer)
        csvirtualServer = {"printName": "Content switching virtual server", "url": "/stat/csvserver", "rootElement": "csvserver", "splitName": "VirtualServer", "limitSplitting": regexCSVirtualServers, "nameAttr": "name", "metrics": []}
        if pollCSVirtualServers:
            endpoints.append(csvirtualServer)
        vpnvirtualServer = {"printName": "VPN virtual server", "url": "/stat/vpnvserver", "rootElement": "vpnvserver", "splitName": "VirtualServer", "limitSplitting": regexVPNVirtualServers, "nameAttr": "name", "metrics": []}
        if pollVPNVirtualServers:
            endpoints.append(vpnvirtualServer)
        interface = {"printName": "Interface", "url": "/stat/interface", "rootElement": "Interface", "splitName": "Interface", "limitSplitting": regexInterfaces, "nameAttr": "id", "metrics": []}
        if pollInterfaces:
            endpoints.append(interface)
        service = {"printName": "Service", "url": "/stat/service", "rootElement": "service", "splitName": "Service", "limitSplitting": regexServices, "nameAttr": "name", "metrics": []}
        if pollServices:
            endpoints.append(service)
        servicegroup = {"printName": "Servicegroup", "url": "/stat/servicegroup", "rootElement": "servicegroup", "splitName": "Servicegroup", "limitSplitting": regexServiceGroups, "nameAttr": "servicegroupname", "metrics": []}
        if pollServiceGroups:
            endpoints.append(servicegroup)
        
        # All the metrics currently polled
        # name = Name of the metric as defined in the plugin.json
        # path = Position in the returned json where the metric value is written
        # metric = absolute, relative, correlation or property
        # expectedValue = If it is a string value instead of a number
        # propSuffix = Append to the end of the captured value when using properties
        system["metrics"].append({"name": "cpu", "path": "cpuusagepcnt", "metric": "absolute", "max": 1000})
        system["metrics"].append({"name": "mgmtcpu", "path": "mgmtcpuusagepcnt", "metric": "absolute", "max": 1000})
        system["metrics"].append({"name": "pktcpu", "path": "pktcpuusagepcnt", "metric": "absolute", "max": 1000})
        system["metrics"].append({"name": "rescpu", "path": "rescpuusagepcnt", "metric": "absolute", "max": 1000})
        system["metrics"].append({"name": "memutil", "path": "memusagepcnt", "metric": "absolute"})
        system["metrics"].append({"name": "memuse", "path": "memuseinmb", "metric": "absolute"})
        system["metrics"].append({"name": "CPU count", "path": "numcpus", "metric": "property"})
        system["metrics"].append({"name": "Start time", "path": "starttime", "metric": "property"})
        system["metrics"].append({"name": "Allocated memory", "path": "hardcoded formula..., memsizemb is not reliable", "metric": "property", "propSuffix": "MB"})
        aaa["metrics"].append({"name": "AAA.totsessiontimeout", "path": "aaatotsessiontimeout", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.authfail", "path": "aaaauthfail", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.authsuccess", "path": "aaaauthsuccess", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.tottmsessions", "path": "aaatottmsessions", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.authnonhttpsuccess", "path": "aaaauthnonhttpsuccess", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.curtmsessions", "path": "aaacurtmsessions", "metric": "absolute"})
        aaa["metrics"].append({"name": "AAA.curicasessions", "path": "aaacuricasessions", "metric": "absolute"})
        aaa["metrics"].append({"name": "AAA.totsessions", "path": "aaatotsessions", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.authonlyhttpfail", "path": "aaaauthonlyhttpfail", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.curicaonlyconn", "path": "aaacuricaonlyconn", "metric": "absolute"})
        aaa["metrics"].append({"name": "AAA.curicaconn", "path": "aaacuricaconn", "metric": "absolute"})
        aaa["metrics"].append({"name": "AAA.authnonhttpfail", "path": "aaaauthnonhttpfail", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.authonlyhttpsuccess", "path": "aaaauthonlyhttpsuccess", "metric": "relative"})
        aaa["metrics"].append({"name": "AAA.cursessions", "path": "aaacursessions", "metric": "absolute"})
        compression["metrics"].append({"name": "httpcomp", "path": "comphttpbandwidthsaving", "metric": "absolute"})
        compression["metrics"].append({"name": "tcpcomp", "path": "comptcpbandwidthsaving", "metric": "absolute"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.hits", "path": "tothits", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vsreq", "path": "totalrequests", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vsresp", "path": "totalresponses", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vssrvconn", "path": "cursrvrconnections", "metric": "absolute"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vscliconn", "path": "curclntconnections", "metric": "absolute"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vshealth", "path": "vslbhealth", "metric": "absolute"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.totreqbytes", "path": "totalrequestbytes", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.totrespbytes", "path": "totalresponsebytes", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.invreqresp", "path": "invalidrequestresponse", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.invreqresdrp", "path": "invalidrequestresponsedropped", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.backuphits", "path": "totvserverdownbackuphits", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.satvsreq", "path": "hardcoded formula...", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.tolvsreq", "path": "toleratingttlbtransactions", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.fruvsreq", "path": "frustratingttlbtransactions", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.actsvc", "path": "actsvcs", "metric": "absolute"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vssurge", "path": "surgecount", "metric": "relative"})
        lbvirtualServer["metrics"].append({"name": "LBVirtualServer.vsstate", "path": "state", "expectedValue": "UP", "metric": "absolute"})
        lbvirtualServer["metrics"].append({"ip": "primaryipaddress", "port": "primaryport", "metric": "correlation"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csstate", "path": "state", "expectedValue": "UP", "metric": "absolute"})
        csvirtualServer["metrics"].append({"ip": "primaryipaddress", "port": "primaryport", "metric": "correlation"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvssatcsvsreq", "path": "hardcoded formula...", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvstolcsvsreq", "path": "toleratingttlbtransactions", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvsfrucsvsreq", "path": "frustratingttlbtransactions", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvshits", "path": "tothits", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvsreq", "path": "totalrequests", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvsresp", "path": "totalresponses", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvssrvconn", "path": "cursrvrconnections", "metric": "absolute"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvscliconn", "path": "curclntconnections", "metric": "absolute"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvstotreqbytes", "path": "totalrequestbytes", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvstotrespbytes", "path": "totalresponsebytes", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvsinvreqresp", "path": "invalidrequestresponse", "metric": "relative"})
        csvirtualServer["metrics"].append({"name": "CSVirtualServer.csvsinvreqresdrp", "path": "invalidrequestresponsedropped", "metric": "relative"})
        vpnvirtualServer["metrics"].append({"name": "VPNVirtualServer.vpnstate", "path": "state", "expectedValue": "UP", "metric": "absolute"})
        vpnvirtualServer["metrics"].append({"ip": "primaryipaddress", "port": "primaryport", "metric": "correlation"})
        vpnvirtualServer["metrics"].append({"name": "VPNVirtualServer.totalrequests", "path": "totalrequests", "metric": "relative"})
        vpnvirtualServer["metrics"].append({"name": "VPNVirtualServer.totalresponses", "path": "totalresponses", "metric": "relative"})
        vpnvirtualServer["metrics"].append({"name": "VPNVirtualServer.totalrequestbytes", "path": "totalrequestbytes", "metric": "relative"})
        vpnvirtualServer["metrics"].append({"name": "VPNVirtualServer.totalresponsebytes", "path": "totalresponsebytes", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.intstate", "path": "curintfstate", "expectedValue": "UP", "metric": "absolute"})
        interface["metrics"].append({"name": "Interface.rxbytes", "path": "totrxbytes", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.txbytes", "path": "tottxbytes", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.errpktrx", "path": "errpktrx", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.errpkttx", "path": "errpkttx", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.discpktrx", "path": "errifindiscards", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.discpkttx", "path": "nicerrifoutdiscards", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.droppktrx", "path": "errdroppedrxpkts", "metric": "relative"})
        interface["metrics"].append({"name": "Interface.droppkttx", "path": "errdroppedtxpkts", "metric": "relative"})
        service["metrics"].append({"name": "Service.svcstate", "path": "state", "expectedValue": "UP", "metric": "absolute"})
        service["metrics"].append({"name": "Service.thput", "path": "throughput", "metric": "absolute"})
        service["metrics"].append({"name": "Service.ttfb", "path": "avgsvrttfb", "metric": "absolute"})
        service["metrics"].append({"name": "Service.svcreq", "path": "totalrequests", "metric": "relative"})
        service["metrics"].append({"name": "Service.svcresp", "path": "totalresponses", "metric": "relative"})
        service["metrics"].append({"name": "Service.reqbytes", "path": "totalrequestbytes", "metric": "relative"})
        service["metrics"].append({"name": "Service.respbytes", "path": "totalresponsebytes", "metric": "relative"})
        service["metrics"].append({"name": "Service.surge", "path": "surgecount", "metric": "relative"})
        service["metrics"].append({"name": "Service.svcsrvconn", "path": "cursrvrconnections", "metric": "absolute"})
        service["metrics"].append({"name": "Service.svccliconn", "path": "curclntconnections", "metric": "absolute"})
        service["metrics"].append({"name": "Service.satsvcreq", "path": "hardcoded formula...", "metric": "relative"})
        service["metrics"].append({"name": "Service.tolsvcreq", "path": "toleratingttlbtransactions", "metric": "relative"})
        service["metrics"].append({"name": "Service.frusvcreq", "path": "frustratingttlbtransactions", "metric": "relative"})
        servicegroup["metrics"].append({"name": "ServiceGroup.sgstate", "path": "state", "expectedValue": "Enabled", "metric": "absolute"})
        
        self.successfullyPolled = False
        if self.api == None:
            self.api = API(logger=logger, username=username, password=password, endpoint=url, verify_ssl=verify, start=self.start, use_token=self.session)
        else:
            self.api.setStart(self.start)
        
        pool = ThreadPool(processes = 10)
        
        for endpoint in endpoints:
            logger.debug("Registering endpoint " + endpoint["printName"])
            pool.apply_async(self.callEndPoint, (self.api, endpoint, e1))
        pool.close()
        
        # Max polling time is 50 seconds
        process_alive = 0
        while time.time() - self.start < 50:
            process_alive = 0
            for process in pool._pool:
                if process.is_alive():
                    process_alive += 1
            if process_alive == 0:
                logger.info('All processes finished, exiting checking loop')
                break
            time.sleep(1)
        
        # If we get to this point, a process was flagged alive and it's been more than 50 seconds, then terminate it
        # Otherwise, if process_alive was not set above, then no process should be alive at this point and no need to terminate
        if process_alive > 0:
            pool.terminate()
            pool.join()
            logger.warning(str(process_alive) + ' processes are alive. Terminating before finishing polling cycle')
            
        # Register all metrics which contain a result
        for endpoint in endpoints:
            for metric in endpoint["metrics"]:
                if "splits" in metric:
                    splits = 0
                    for split in metric["splits"]:
                        if "max" in metric and split["result"] > metric["max"]:
                            logger.debug("Value too high (" + str(split["result"]) + "), skipping " + metric["name"] + " with splittings " + str(split))
                        else:
                            if splits < self.MAX_SPLITS:
                                if metric["metric"] == "absolute":
                                    e1.absolute(key = metric["name"], value = split["result"], dimensions = split["dimensions"])
                                else:
                                    e1.relative(key = metric["name"], value = split["result"], dimensions = split["dimensions"])
                                splits += 1
                            else:
                                logger.debug("Max of " + str(self.MAX_SPLITS) + " splittings reached, skipping " + metric["name"] + " with splittings " + str(split))
                elif "result" in metric:
                    if "max" in metric and metric["result"] > metric["max"]:
                        logger.debug("Value too high (" + str(metric["result"]) + "), skipping " + metric["name"])
                    else:
                        if metric["metric"] == "absolute":
                            e1.absolute(key = metric["name"], value = metric["result"])
                        else:
                            e1.relative(key = metric["name"], value = metric["result"])
                        
        if self.successfullyPolled:
            e1.absolute(key = "connectivity", value = 100)
        else:
            e1.absolute(key = "connectivity", value = 0)
            
        hostNameIsIP = False
        try:
            socket.inet_aton(hostName)
            hostNameIsIP = True
        except socket.error as e:
            hostNameIsIP = False

        if hostNameIsIP:
            self.listeningIPs.add(hostName)
        
        endpointsAssigned = 0
        for ip in self.listeningIPs:
            if ip != "0.0.0.0" and ip != "127.0.0.1":
                # Check for valid IP
                ipIsValid = False
                try:
                    socket.inet_aton(ip)
                    ipIsValid = True
                except socket.error as e:
                    ipIsValid = False
                if ipIsValid:
                    try:
                        # Version 163+ allows for an endpoint without port
                        if endpointsAssigned < 100:
                            e1.add_endpoint(ip)
                            logger.debug("Adding endpoint " + ip)
                            endpointsAssigned += 1
                        else:
                            logger.info("Too many endpoints, skipping " + ip)
                    except TypeError as e:
                        # Fallback for older versions
                        for port in self.listeningPorts:
                            if not isinstance(port, int) and port.isdigit():
                                port = int(port)
                            if port > 0 and port < 65536:
                                if endpointsAssigned < 100:
                                    e1.add_endpoint(ip, int(port))
                                    logger.debug("Adding endpoint " + ip + ":" + str(port))
                                    endpointsAssigned += 1
                                else:
                                    logger.info("Too many endpoints, skipping " + ip + ":" + str(port))
            
        self.previousEvents = self.raisedEvents
    
    def callEndPoint(self, api, endpoint, e1):
        try:
            response = api.request(endpoint["url"])
            if response is None:
                logger.info("No '" + endpoint["printName"] + "' statistics received, cannot fetch metrics")
            else:
                if endpoint["rootElement"] in response:
                    self.successfullyPolled = True
                    stats = response[endpoint["rootElement"]]
                    if "splitName" in endpoint:
                        for element in stats:
                            self.getMetrics(endpoint, element, e1)
                    else:
                        self.getMetrics(endpoint, stats, e1)
                else:
                    logger.info("No '" + endpoint["printName"] + "' configured, cannot fetch metrics")
        except Exception as e:
            logger.info("Failed polling endpoint: " + str(endpoint))
            logger.exception(e)

    # Extract metrics from responses
    def getMetrics(self, endpoint, stats, e1):
        if "nameAttr" in endpoint:
            name = stats[endpoint["nameAttr"]]
        else:
            name = ""
        for metric in endpoint["metrics"]:
            if metric["metric"] == "correlation":
                self.listeningPorts.add(stats[metric["port"]])
                self.listeningIPs.add(stats[metric["ip"]])
            elif metric["metric"] == "property":
                if metric["name"] == "Allocated memory":
                    value = int(int(stats["memuseinmb"]) / (int(stats["memusagepcnt"]) / 100))
                else:
                    value = stats[metric["path"]]
                suffix = ""
                if "propSuffix" in metric:
                    suffix = metric["propSuffix"]
                e1.report_property(metric["name"], str(value) + suffix)
            else:
                if metric["name"] == "satvsreq" or metric["name"] == "csvssatcsvsreq":
                    if "totcltttlbtransactions" in stats and "toleratingttlbtransactions" in stats and "frustratingttlbtransactions" in stats:
                        value = int(stats["totcltttlbtransactions"]) - int(stats["toleratingttlbtransactions"]) - int(stats["frustratingttlbtransactions"])
                    else:
                        value = None
                elif metric["name"] == "satsvcreq":
                    if "totsvrttlbtransactions" in stats and "toleratingttlbtransactions" in stats and "frustratingttlbtransactions" in stats:
                        value = int(stats["totsvrttlbtransactions"]) - int(stats["toleratingttlbtransactions"]) - int(stats["frustratingttlbtransactions"])
                    else:
                        value = None
                elif metric["path"] in stats:
                    value = stats[metric["path"]]
                else:
                    value = None
                if "expectedValue" in metric:
                    if metric["expectedValue"].lower() == value.lower():
                        value = 100
                    else:
                        self.report_custom_info_event(e1, endpoint["printName"] + " '" + name + "' is not '" + metric["expectedValue"] + "', it is '" + value + "'", endpoint["printName"] + " unavailable")
                        value = 0
                dimensions = {}
                if "splitName" in endpoint:
                    dimensions[endpoint["splitName"]] = name
                if name == "":
                    metric["result"] = value
                else:
                    splitLimitationPassed = False
                    if "limitSplitting" in endpoint:
                        if len(endpoint["limitSplitting"]) == 0:
                            splitLimitationPassed = True
                            logger.info("Matching splitting due to 0 size filter: " + name)
                        else:
                            for regex in endpoint["limitSplitting"]:
                                if regex.match(name):
                                    splitLimitationPassed = True
                                    logger.info("Matching splitting due to regex '" + str(regex) + "': " + name)
                                    break
                        if not splitLimitationPassed:
                            logger.info("Limiting splitting, text: " + name)
                    else:
                        splitLimitationPassed = True
                    if splitLimitationPassed:
                        if not "splits" in metric:
                            metric["splits"] = []
                        if value is not None:
                            metric["splits"].append({"result": value, "dimensions": dimensions})
                    
    def report_custom_info_event(self, e1, message, title):
        if self.events < self.MAX_EVENTS:
            if title + message not in self.previousEvents:
                self.events += 1
                e1.report_custom_info_event(message, title)
            self.raisedEvents.append(title + message)
            
import json
import requests
import urllib
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

class API(object):
    def __init__(
            self, 
            username,
            password,
            endpoint,
            base_path="/nitro/v1",
            logger=None,
            verify_ssl=True,
            start=0,
            use_token=False):

        self.username = username
        self.password = password
        self.endpoint = endpoint
        self.base_path = base_path
        self.logger = logger
        self.verify_ssl = verify_ssl
        self.start = start
        self.use_token = use_token
        self.token = None
        self.createSession()
        
    def setStart(self, start):
        self.start = start
        
    def logout(self):
        if self.use_token:
            payload = {"object": json.dumps({"logout": {}})}
            self.request('/config/logout', payload)
        self.session = None
        self.token = None
        
    def createSession(self):
        self.session = requests.Session()
        retry = Retry(connect=3, backoff_factor=0.5)
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)
        if self.use_token:
            payload = {"object": json.dumps({"login": {"username": self.username, "password": self.password}})}
            self.request('/config/login', payload)
        
    def request(self, path, payload=None):
        if self.session == None:
            self.createSession()
        result = None
        headers = {'Accept':'application/json', 'Connection':'keep-alive'}
        if not self.use_token:
            headers['X-NITRO-USER'] = self.username
            headers['X-NITRO-PASS'] = self.password
            
        cookies = {}
        if self.token:
            cookies['NITRO_AUTH_TOKEN'] = self.token
            self.logger.info("NITRO_AUTH_TOKEN: " + self.token)
            
        if not self.use_token or (self.token or (self.username and self.password and payload and 'login' in payload["object"])):
            url = '%s/%s/%s' % (self.endpoint.rstrip('/'), self.base_path.strip('/'), path.strip('/'))
            self.logger.info("Polling URL: " + url)

            try:
                if payload:
                    headers['Content-Type'] = "application/x-www-form-urlencoded"
                    response = self.session.post(url, data=urllib.parse.urlencode(payload), headers=headers, cookies=cookies, verify=self.verify_ssl, timeout=48 - (time.time() - self.start))
                else:
                    response = self.session.get(url, headers=headers, cookies=cookies, verify=self.verify_ssl, timeout=48 - (time.time() - self.start))
                if response.ok and payload and 'login' in payload["object"]:
                    self.logger.info("Login")
                    self.token = response.cookies['sessionid']
                    if response.text == '':
                        result = {'headers': dict(response.headers)}
                    else:
                        result = {'result': response.text}
                else:
                    if response.ok:
                        try:
                            result = dict(response.json())
                        except:
                            if response.text == '':
                                result = {"headers": dict(response.headers)}
                            else:
                                result = {"result": response.text}
                    else:
                        self.logger.error("URL: " + url + ". Response: " + response.text + ". Restarting session.")
                        self.logout()
            except Exception as e:
                logger.warn("Failed polling URL: " + str(url) + ". Restarting session.")
                logger.exception(e)
                self.logout()
            return result
        else:
            self.logger.error("Endpoint, username and password are required to use the api.")
            self.session = None
            return None